package buildme

import (
	"bufio"
	"buildme/utils"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"

	"dev.hexasoftware.com/hxs/prettylog"

	"github.com/docker/docker/api/types"
	"github.com/docker/docker/api/types/container"
	docker "github.com/docker/docker/client"
)

type Task struct {
	Name string
	Conf TaskConf // Current conf
	proj *Project
	// Use

	netID string
	imgID string
}

func (t *Task) Exec() error {
	// Create network
	cli, err := docker.NewEnvClient()
	if err != nil {
		return err
	}
	defer cli.Close()
	/*
		ctx := context.Background()
			networkName := t.proj.Conf.Project + "-" + t.Name
			netInspect, err := cli.NetworkInspect(ctx, networkName, false)
			if err == nil {
				log.Println("======= Network found")
				t.netID = netInspect.ID
			} else {
				log.Println("======= Creating network")
				netRes, err := cli.NetworkCreate(ctx, networkName, types.NetworkCreate{})
				if err != nil {
					return err
				}
				t.netID = netRes.ID
			}*/

	if err := t.startServices(); err != nil {
		return err
	}
	if t.Conf.Dockerfile != "" {
		if err := t.buildImage(); err != nil {
			return err
		}
	} else if t.Conf.Image != "" {
		log.Println("Using image:", t.Conf.Image)
		t.imgID = t.Conf.Image

	} else {
		errors.New("Image or dockerfile entry not found")
	}

	if err := t.processCommands(); err != nil {
		return err
	}
	return nil
}
func (t *Task) startService(name string) error {
	log.Println("Starting service:", name)
	cli, err := docker.NewEnvClient()
	if err != nil {
		return err
	}
	defer cli.Close()
	ctx := context.Background()

	srv, ok := t.proj.Conf.Services[name]
	if !ok {
		return fmt.Errorf("Service not found '%s'", name)
	}

	serviceName := t.proj.Conf.Project + "-" + name
	log.Println("Start container with:", serviceName)

	im, _, err := cli.ImageInspectWithRaw(ctx, srv.Image)
	if err != nil {
		return nil

	}

	conInspect, err := cli.ContainerInspect(ctx, serviceName)
	if err == nil {
		if conInspect.State.Running {
			log.Println("Recreate ", serviceName)
			if err := cli.ContainerKill(ctx, serviceName, "KILL"); err != nil {
				return err
			}
		}
		if err := cli.ContainerRemove(ctx, serviceName, types.ContainerRemoveOptions{RemoveVolumes: true}); err != nil {
			return err
		}
	}

	hostConfig := &container.HostConfig{}
	log.Println("Container create with opts:", im.Config)
	conConfig := &container.Config{Image: srv.Image, Cmd: im.Config.Cmd, Hostname: name}
	con, err := cli.ContainerCreate(ctx, conConfig, hostConfig, nil, serviceName)
	if err != nil {
		return err
	}
	/*log.Println("Network connect")
	err = cli.NetworkConnect(ctx, t.netID, con.ID, &network.EndpointSettings{})
	if err != nil {
		return err
	}*/

	log.Println("Start")
	err = cli.ContainerStart(ctx, con.ID, types.ContainerStartOptions{})
	if err != nil {
		return err
	}

	return nil
}

func (t *Task) startServices() error {
	for _, v := range t.Conf.Services {
		// Start service
		err := t.startService(v)
		if err != nil {
			return err
		}
	}
	return nil
}

// only if docker file
func (t *Task) buildImage() error {
	cli, err := docker.NewEnvClient()
	if err != nil {
		return err
	}
	defer cli.Close()
	ctx := context.Background()
	log.Println("Using image:", t.Conf.Dockerfile)

	dockerTag := t.proj.Conf.Project + "-" + t.Name
	// Execute inside container
	log.Println("===== Building IMAGE")
	imgBuildOptions := types.ImageBuildOptions{
		Tags:       []string{dockerTag},
		Dockerfile: t.Conf.Dockerfile,
		Remove:     true,
	}
	res, err := cli.ImageBuild(ctx, t.proj.Reader(), imgBuildOptions)
	if err != nil {
		return err
	}
	defer res.Body.Close()

	// Just a reader and output
	br := bufio.NewReader(res.Body)
	for {
		jdata, _, err := br.ReadLine()
		if err != nil {
			break
		}
		mdata := map[string]interface{}{}
		json.Unmarshal(jdata, &mdata)

		msg, ok := mdata["stream"]
		if !ok {
			msg = mdata["error"] // Error if exists
		}
		if msg != nil {
			log.Println(msg)
		}
	}
	t.imgID = dockerTag
	// Just output
	return nil
}

func (t *Task) processCommands() error {
	log.Println("Executing commands")

	img, err := ImageCli(t.imgID)
	if err != nil {
		return err
	}
	// Builds image
	for _, cmd := range t.Conf.Command {
		vcmd := utils.ParseField(cmd)
		var err error
		switch vcmd[0] {
		case "RUN":
			err = t.Run(vcmd[1:])
		case "GET":
			err = img.Get(vcmd[1], vcmd[2])
		}
		if err != nil {
			return err
		}
	}
	return nil

}

// Specific for docker?
func (t *Task) Run(command []string) error {
	log := prettylog.New("RUN in IMAGE:")
	log.Println("RUN", command)

	cli, err := docker.NewEnvClient()
	if err != nil {
		return err
	}
	defer cli.Close()
	ctx := context.Background()

	conConfig := &container.Config{
		Image:    t.imgID,
		Cmd:      command,
		Env:      t.Conf.Env,
		Hostname: t.Name,
	}
	// Mount links
	links := []string{}

	for _, v := range t.Conf.Services {
		links = append(links, t.proj.Conf.Project+"-"+v+":"+v)
		log.Println("Linking to:", links)
	}
	hostConfig := &container.HostConfig{Links: links}
	ncon, err := cli.ContainerCreate(ctx, conConfig, hostConfig, nil, "")
	if err != nil {
		return err
	}
	defer cli.ContainerRemove(ctx, ncon.ID, types.ContainerRemoveOptions{RemoveVolumes: true})

	/*log.Println("Network connect")
	err = cli.NetworkConnect(ctx, t.netID, ncon.ID, &network.EndpointSettings{Links: []string{"mongodb"}})
	if err != nil {
		return err
	}*/

	//c.conID = ncon.ID
	err = cli.ContainerStart(ctx, ncon.ID, types.ContainerStartOptions{})
	if err != nil {
		return err
	}
	res, err := cli.ContainerAttach(ctx, ncon.ID, types.ContainerAttachOptions{Stream: true, Stdout: true, Stderr: true, Logs: true})
	if err != nil {
		return err
	}
	io.Copy(prettylog.NewWriter("OUT"), res.Conn)
	retCode, err := cli.ContainerWait(ctx, ncon.ID)
	if err != nil {
		return err
	}
	_ = retCode
	/*if retCode != 0 {
		return errors.New("Container did not complete successfully")
	}*/

	imgRes, err := cli.ContainerCommit(ctx, ncon.ID, types.ContainerCommitOptions{Config: conConfig})

	cli.ImageTag(ctx, imgRes.ID, t.imgID+":building")
	return nil
}
